'use strict'

const fetch = require('node-fetch');
const Signer = require('./../util/signer');
const Factory = require("./factory");
const Config = require("../config");
const Clock = require("../util/clock");
const Math = require("../util/math");
const CODE = require("../code");
const Hasher = require("../util/hasher");

module.exports = class Sender {

    host = null;
    private_key = null;

    constructor(host = null, private_key = null) {
        try {
            this.Host(host);
            this.PrivateKey(private_key);
        } catch (e) {
            throw e;
        }
    }

    Host(host = null) {
        this.host = host ?? this.host;

        return this.host;
    }

    PrivateKey(private_key = null) {
        this.private_key = private_key ?? this.private_key;
        if(this.private_key === null) this.private_key = Signer.PrivateKey();

        return this.private_key;
    }

    async RPC(rpc, data = {}) {
        try {
            return await fetch(`${this.host}/${rpc}`, {
                method: 'POST',
                headers: {
                    'Content-Type': 'application/json;charset=utf-8'
                },
                body: JSON.stringify(data),

            })
                .then(res => res.json())
                .then(res => {
                    if(res.code && res.code !== 200) throw res;
                    return res;
                })
                .catch(e => {
                    throw e;
                });
        } catch (e) {
            throw e;
        }
    }

    Request(request = {}, public_key = '', signature = '') {
        try {
            if(typeof request === 'object') request = JSON.stringify(request);

            return this.RPC('request', {
                request: request,
                public_key: public_key,
                signature: signature,
            });
        } catch (e) {
            throw e;
        }
    }

    RawRequest(request = {}, public_key = '', signature = '') {
        try {
            return this.RPC('rawrequest', {
                request: request,
                public_key: public_key,
                signature: signature,
            });
        } catch (e) {
            throw e;
        }
    }

    SendTransaction(transaction = {}, public_key = '', signature = '') {
        try {
            if(typeof transaction === 'object') transaction = JSON.stringify(transaction);
            return this.RPC('sendtransaction', {
                transaction: transaction,
                public_key: public_key,
                signature: signature,
            });

        } catch (e) {
            throw e;
        }
    }

    SendTransactionWithResult(signedTransaction, waitMicroSeconds = 60000000, blockSearchPadding = 10) {
        try {
            return this.SendTransaction(signedTransaction.data, signedTransaction.public_key, signedTransaction.signature).then(async res => {
                if(res.code !== CODE.OK || res.status !== 'success' || res.msg !== 'Transaction is added') throw res;

                console.log(res);

                return await this.Checker.TransactionResult(signedTransaction, waitMicroSeconds, blockSearchPadding).catch(e => {
                    throw e;
                });
            }).catch(e => {
                throw e;
            });

        } catch (e) {
            throw e;
        }
    }

    SendRawTransaction(transaction = {}, public_key = '', signature = '') {
        try {
            return this.RPC('sendrawtransaction', {
                transaction: transaction,
                public_key: public_key,
                signature: signature,
            });
        } catch (e) {
            throw e;
        }
    }

    Broadcast(chain_type='main', round_key='') {
        try {
            return this.RPC('broadcast',{
                chain_type: chain_type,
                round_key: round_key
            });
        } catch (e) {
            throw e;
        }
    }

    Info() {
        try {
            return this.RPC('info');
        } catch (e) {
            throw e;
        }
    }

    Main() {
        try {
            return this.RPC('main');
        } catch (e) {
            throw e;
        }
    }

    Peer(prefix='', register=0, host='', authentication=0, height=0) {
        try {
            return this.RPC('peer',{
                prefix: prefix,
                register: register,
                host: host,
                authentication: authentication,
                height: height
            });
        } catch (e) {
            throw e;
        }
    }

    Ping() {
        try {
            return this.RPC('ping');
        } catch (e) {
            throw e;
        }
    }

    Round(chain_type='main',height=null) {
        try {
            return this.RPC('round',{
                chain_type: chain_type,
                height: height
            });
        } catch (e) {
            throw e;
        }
    }

    Status(signed_query={}) {
        try {
            return this.RPC('status',{
                signed_query: signed_query
            });
        } catch (e) {
            throw e;
        }
    }

    Checker = {
        /**
         * Get byte fee of the contract
         * @param contractType : string
         * @param blockSearchStart : number
         * @param blockSearchLimit : number
         * @returns {Promise<void>}
         * @constructor
         */
        ContractFee: async (contractType, blockSearchStart, blockSearchLimit = 1000) => {
            try {
                const latestBlock = await this.Round();
                let searchBlockTarget = latestBlock.data.block.height;
                let searchBlockTargetHeight = latestBlock.data.block.height;
                let feeAmount = null;
                let examineData = {
                    block: null,
                    transaction: null,
                };
                if(blockSearchStart) {
                    searchBlockTarget = blockSearchStart;
                    searchBlockTargetHeight = blockSearchStart;
                }

                // get contract code
                const ctype = 'contract';
                const cid = Hasher.IdHash([contractType, Config.SystemNonce()]);
                const contractIndex = Config.ContractPrefix() + cid;
                const signedRequestGetCode = Factory.Request('GetCode', {target: cid, ctype: ctype}, this.private_key);
                const getCode = await this.Request(
                    signedRequestGetCode.data,
                    signedRequestGetCode.public_key,
                    signedRequestGetCode.signature
                );

                if(getCode.data === undefined || getCode.data === null) throw `GetCode "${contractType}" failed. `;
                if(getCode.data[contractIndex] === undefined || getCode.data[contractIndex] === null) throw `Contract "${contractType}" doesn\'t exist. `;

                const contract = JSON.parse(getCode.data[contractIndex]);

                while(examineData.block === null && examineData.transaction === null && searchBlockTargetHeight < blockSearchLimit + searchBlockTarget) {
                    const signedRequestGetBlock = Factory.Request('GetBlock', {target: searchBlockTarget, full: true}, this.private_key);
                    const getBlock = await this.Request(
                        signedRequestGetBlock.data,
                        signedRequestGetBlock.public_key,
                        signedRequestGetBlock.signature
                    ).catch(e => {
                        return {};
                    });

                    if(getBlock.data && getBlock.data.transactions) {
                        const txHashs = Object.keys(getBlock.data.transactions);
                        txHashs.forEach(hash => {
                            const transaction = getBlock.data.transactions[hash];
                            if(transaction.transaction.type === contract.name) {
                                examineData.block = getBlock.data;
                                examineData.transaction = transaction;
                            }
                        });
                    }
                    console.log(searchBlockTarget);
                    searchBlockTarget--;
                }

                if(examineData.block === null || examineData.transaction === null) throw 'No block or transaciton found.';

                let balanceUpdates = {};
                contract.updates.forEach(update => {
                    let key, value;
                    if(update.hasOwnProperty('$write_universal') && update['$write_universal'][0] === 'balance') {
                        key = update['$write_universal'][1];
                        value = update['$write_universal'][2];
                        if(typeof key === 'object' && key.hasOwnProperty('$load_param')) key = key['$load_param'][0];
                        balanceUpdates[key] = value;
                    }
                });

                if(balanceUpdates.hasOwnProperty('from')) feeAmount = 0;

                for (const balanceUpdatesKey in balanceUpdates) {
                    const balanceStatusHash = Hasher.StatusHash(contract.writer, contract.nonce, 'balance', examineData.transaction.transaction[balanceUpdatesKey]);
                    const oldBalance = examineData.block.universal_updates[balanceStatusHash].old;
                    const newBalance = examineData.block.universal_updates[balanceStatusHash].new;
                    if(balanceUpdatesKey === 'from') feeAmount += Math.Sub(oldBalance, newBalance);
                    else feeAmount -= Math.Sub(newBalance, oldBalance);
                }

                if(feeAmount === null) return feeAmount;
                else return Math.Div(feeAmount, JSON.stringify(examineData.transaction).length.toString(), 0);

            } catch (e) {
                throw e;
            }

        },
        TransactionResult: (signedTransaction, waitMicroSeconds = Config.CONFIRM_INTERVAL, blockSearchPadding = 10) => {
            try {
                if(!Number.isInteger(blockSearchPadding)) throw `blockSearchPadding must be an integer.`;
                if(!Number.isInteger(waitMicroSeconds)) throw `waitMicroSeconds must be an integer.`;
                if(!Number.isInteger(signedTransaction.timestamp)) throw `signedTransaction.timestamp must be an integer.`;
                if(!Hasher.IsHex(signedTransaction.hash)) throw `signedTransaction.hash must be a hex string.`;

                const recursiveCheckResult = async () => {
                    try {
                        const status = {
                            pending: 'pending',
                            reject: 'reject',
                            accept: 'accept',
                        };
                        const latestBlock = await this.Round();
                        if(latestBlock.data.block.s_timestamp < signedTransaction.timestamp) return { status: status.pending };
                        else {
                            const signedRequestGetTransaction = Factory.Request('GetTransaction', {target: signedTransaction.hash}, this.private_key);
                            const getTransaction = await this.Request(
                                signedRequestGetTransaction.data,
                                signedRequestGetTransaction.public_key,
                                signedRequestGetTransaction.signature
                            ).catch(e => {
                                return {};
                            });
                            let data = getTransaction.data || {};
                            if(data.transaction === undefined || data.transaction === null) data = {};
                            data.hash = signedTransaction.hash;
                            data.block = null;

                            if(data.transaction !== undefined && data.transaction !== null){
                                let searchBlockTarget = latestBlock.data.block.height + blockSearchPadding;

                                while(data.block === null && searchBlockTarget + blockSearchPadding >= latestBlock.data.block.height) {
                                    const signedRequestGetBlock = Factory.Request('GetBlock', {target: searchBlockTarget, full: true}, this.private_key);
                                    const getBlock = await this.Request(
                                        signedRequestGetBlock.data,
                                        signedRequestGetBlock.public_key,
                                        signedRequestGetBlock.signature
                                    ).catch(e => {
                                        return {};
                                    });
                                    if(
                                        getBlock.data &&
                                        getBlock.data.transactions &&
                                        getBlock.data.transactions.hasOwnProperty(signedTransaction.hash)
                                    ) {
                                        data.block = getBlock.data;
                                    } else searchBlockTarget--;
                                }
                                data.status = status.accept;
                            } else {
                                data.status = status.reject;
                            }

                            return data;
                        }
                    } catch (e) {
                        throw e;
                    }
                }

                return new Promise(async (resolve, reject) => {
                    try {
                        let transactionResult = { status: 'pending' };
                        const limit = Clock.Utime() + Number(waitMicroSeconds);
                        while(transactionResult.status === 'pending' && Clock.Utime() < limit) {
                            transactionResult = await recursiveCheckResult();
                            if(transactionResult.status !== 'pending') break;
                        }
                        if(transactionResult.status !== undefined && transactionResult.status !== null) resolve(transactionResult);
                        else reject(transactionResult);
                    } catch (e) {
                        throw e;
                    }
                });
            } catch (e) {
                throw e;
            }
        }
    }
}